/*
 * PostgreSQL functions to interface with memcache.
 *
 * Copyright (c) 2004-2005 Sean Chittenden <sean@chittenden.org>
 * Copyright (c) 2007 Neil Conway <neilc@samurai.com>
 * Copyright (c) 2007 Open Technology Group, Inc. <http://www.otg-nc.com>
 *
 * See the file COPYING for distribution terms.
 */

#include "postgres.h"

#include <inttypes.h>

#include "fmgr.h"
#include "lib/stringinfo.h"
#include "utils/builtins.h"
#include "utils/datetime.h"
#include "utils/memutils.h"

#include "pgmemcache.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

/* Global memcache instance. */
static struct memcache *mc = NULL;
static struct memcache_ctxt *ctxt = NULL;

/* Add a small work around function to compensate for PostgreSQLs lack
 * of actual memory functions: PostgreSQL supplies macros instead. */
static void	 mcm_pfree(void *ptr);
static void	*mcm_palloc(const size_t size);
static void	*mcm_repalloc(void *ptr, const size_t size);
static Datum  memcache_atomic_op(int type, PG_FUNCTION_ARGS);
static text	 *memcache_gen_host(const struct memcache_server *ms);
static Datum  memcache_set_cmd(int type, PG_FUNCTION_ARGS);

#define MCM_SET_CMD_TYPE_ADD            0x0001
#define MCM_SET_CMD_TYPE_REPLACE        0x0002
#define MCM_SET_CMD_TYPE_SET            0x0004

#define MCM_DATE_TYPE_INTERVAL          0x0010
#define MCM_DATE_TYPE_TIMESTAMP         0x0020

#define MCM_ATOMIC_TYPE_DECR            0x0040
#define MCM_ATOMIC_TYPE_INCR            0x0080

/* Don't add a server until a caller has called memcache_init().  This
 * is probably an unnecessary degree of explicitness, but it forces
 * good calling conventions for pgmemcache.  */
#define MCM_CHECK(_ret) do { \
  if (mc == NULL || ctxt == NULL) { \
    elog(ERROR, "%s(): mc is NULL, call memcache_init()", __FUNCTION__); \
    _ret; \
  } \
} while(0)

/* Add a small work around function to compensate for PostgreSQLs lack
 * of a palloc, pfree, and repalloc functions.  They exist, but are
 * macros, which is useless with regards to function pointers if the
 * macro doesn't have the same signature.  Create that signature
 * here. */
static void
mcm_pfree(void *ptr) {
  return pfree(ptr);
}


static void *
mcm_palloc(const size_t size) {
  return palloc(size);
}


static void *
mcm_repalloc(void *ptr, const size_t size) {
  return repalloc(ptr, size);
}

Datum
memcache_add(PG_FUNCTION_ARGS) {
  return memcache_set_cmd(MCM_SET_CMD_TYPE_ADD | MCM_DATE_TYPE_INTERVAL, fcinfo);
}


Datum
memcache_add_absexpire(PG_FUNCTION_ARGS) {
  return memcache_set_cmd(MCM_SET_CMD_TYPE_ADD | MCM_DATE_TYPE_TIMESTAMP, fcinfo);
}


static Datum
memcache_atomic_op(int type, PG_FUNCTION_ARGS) {
  text		*key = PG_GETARG_TEXT_P(0);
  size_t	 key_len;
  u_int32_t	 val,
		 ret = 0;

  MCM_CHECK(PG_RETURN_NULL());

  key_len = VARSIZE(key) - VARHDRSZ;
  if (key_len < 1)
    elog(ERROR, "unable to have a zero length key");

  val = 1;
  if (PG_NARGS() >= 2 && PG_ARGISNULL(1) == false)
    val = PG_GETARG_UINT32(1);

  if (type & MCM_ATOMIC_TYPE_DECR)
    ret = mcm_decr(ctxt, mc, VARDATA(key), key_len, val);
  else if (type & MCM_ATOMIC_TYPE_INCR)
    ret = mcm_incr(ctxt, mc, VARDATA(key), key_len, val);
  else
    elog(ERROR, "%s():%s:%u\tunknown atomic type 0x%x",
         __FUNCTION__, __FILE__, __LINE__, type);

  PG_RETURN_UINT32(ret);
}


Datum
memcache_decr(PG_FUNCTION_ARGS) {
  return memcache_atomic_op(MCM_ATOMIC_TYPE_DECR, fcinfo);
}


Datum
memcache_delete(PG_FUNCTION_ARGS) {
  text		*key = PG_GETARG_TEXT_P(0);
  size_t	 key_len;
  float8	 hold;
  int		 ret;

  MCM_CHECK(PG_RETURN_NULL());

  key_len = VARSIZE(key) - VARHDRSZ;
  if (key_len < 1)
    elog(ERROR, "unable to have a zero length key");

  hold = 0.0;
  if (PG_NARGS() >= 2 && PG_ARGISNULL(1) == false) {
    Interval *span = PG_GETARG_INTERVAL_P(1);

#ifdef HAVE_INT64_TIMESTAMP
    hold = (span->time / 1000000e0);
#else
    hold = span->time;
#endif
    if (span->month != 0) {
      hold += ((365.25 * 86400) * (span->month / 12));
      hold += ((30.0 * 86400) * (span->month % 12));
    }
  }

  ret = mcm_delete(ctxt, mc, VARDATA(key), key_len, (time_t) hold);

  if (ret < 0)
    elog(ERROR, "internal libmemcache(3) error");

  PG_RETURN_BOOL(ret == 0);
}


/* Deprecated: use memcache_flush() */
Datum
memcache_flush_all(PG_FUNCTION_ARGS) {
  return memcache_flush(fcinfo);
}


Datum
memcache_flush_all0(PG_FUNCTION_ARGS) {
  text		*key = PG_GETARG_TEXT_P(0);
  size_t	 key_len;
  int		 ret;

  MCM_CHECK(PG_RETURN_NULL());

  key_len = VARSIZE(key) - VARHDRSZ;
  if (key_len < 1)
    elog(ERROR, "unable to have a zero length key");

  ret = mcm_flush_all(ctxt, mc);

  if (ret == 0)
    PG_RETURN_BOOL(true);
  else if (ret > 0)
    PG_RETURN_BOOL(false);
  else
    elog(ERROR, "internal libmemcache(3) error");

  PG_RETURN_BOOL(ret);
}


Datum
memcache_flush(PG_FUNCTION_ARGS) {
  struct memcache_server *ms;
  text		*key;
  size_t	 key_len;
  int		 ret;
  u_int32_t	 hash;

  MCM_CHECK(PG_RETURN_NULL());

  key = PG_GETARG_TEXT_P(0);
  key_len = VARSIZE(key) - VARHDRSZ;
  if (key_len < 1)
    elog(ERROR, "unable to have a zero length key");

  hash = mcm_hash_key(ctxt, VARDATA(key), key_len);
  ms = mcm_server_find(ctxt, mc, hash);

  ret = mcm_flush(ctxt, mc, ms);

  if (ret == 0)
    PG_RETURN_BOOL(true);
  else if (ret > 0)
    PG_RETURN_BOOL(false);
  else
    elog(ERROR, "internal libmemcache(3) error");

  PG_RETURN_BOOL(ret);
}


Datum
memcache_free(PG_FUNCTION_ARGS) {
  MCM_CHECK(PG_RETURN_BOOL(false));

  mcm_free(ctxt, mc);
  mc = NULL;

  mcMemFreeCtxt(ctxt);
  ctxt = NULL;

  PG_RETURN_BOOL(true);
}


Datum
memcache_get(PG_FUNCTION_ARGS) {
  text			      *key = PG_GETARG_TEXT_P(0);
  text			      *ret;
  size_t		       key_len;
  struct memcache_req *req;
  struct memcache_res *res;

  MCM_CHECK(PG_RETURN_NULL());

  key_len = VARSIZE(key) - VARHDRSZ;
  if (key_len < 1)
    elog(ERROR, "unable to have a zero length key");

  req = mcm_req_new(ctxt);
  if (req == NULL)
    elog(ERROR, "unable to create a new memcache request structure");

  /*
   * Note that we pass the content of the key's data section directly
   * to libmemcache. We could call textout(), but we don't need a
   * NUL-terminated string, and textout() does a palloc(), which we'd
   * rather avoid.
   */
  res = mcm_req_add(ctxt, req, VARDATA(key), key_len);
  if (res == NULL)
    elog(ERROR, "unable to create a new memcache response structure");

  mcm_get(ctxt, mc, req);

  if (!(res->_flags & MCM_RES_FOUND))
    PG_RETURN_NULL();

  ret = (text *) palloc(res->bytes + VARHDRSZ);
  VARATT_SIZEP(ret) = res->bytes + VARHDRSZ;
  memcpy(VARDATA(ret), res->val, res->bytes);

  PG_RETURN_TEXT_P(ret);
}


Datum
memcache_hash(PG_FUNCTION_ARGS) {
  text		*key = PG_GETARG_TEXT_P(0);
  size_t	 key_len;
  u_int32_t	 hash;

  key_len = VARSIZE(key) - VARHDRSZ;
  if (key_len < 1)
    elog(ERROR, "unable to have a zero length key");

  hash = mcm_hash_key(ctxt, VARDATA(key), key_len);
  PG_RETURN_UINT32(hash);
}


Datum
memcache_incr(PG_FUNCTION_ARGS) {
  return memcache_atomic_op(MCM_ATOMIC_TYPE_INCR, fcinfo);
}

Datum
memcache_init(PG_FUNCTION_ARGS) {
  MemoryContext	oc;

  /* Return FALSE that way callers can conditionally add servers to
   * the instance. */
  if (ctxt != NULL)
    PG_RETURN_BOOL(false);

  /* Make sure mc is allocated in the top memory context */
  oc = MemoryContextSwitchTo(TopMemoryContext);

  /* Initialize libmemcache's memory functions */
  ctxt = mcMemNewCtxt(mcm_pfree, mcm_palloc, NULL, mcm_repalloc);
  if (ctxt == NULL)
    elog(ERROR, "memcache_init: unable to create a memcache(3) memory context");

  mc = mcm_new(ctxt);
  if (mc == NULL)
    elog(ERROR, "memcache_init: unable to allocate memory");

  /* Return to old memory context */
  MemoryContextSwitchTo(oc);

  PG_RETURN_BOOL(true);
}


Datum
memcache_replace(PG_FUNCTION_ARGS) {
  return memcache_set_cmd(MCM_SET_CMD_TYPE_REPLACE | MCM_DATE_TYPE_INTERVAL, fcinfo);
}


Datum
memcache_replace_absexpire(PG_FUNCTION_ARGS) {
  return memcache_set_cmd(MCM_SET_CMD_TYPE_REPLACE | MCM_DATE_TYPE_TIMESTAMP, fcinfo);
}


Datum
memcache_set(PG_FUNCTION_ARGS) {
  return memcache_set_cmd(MCM_SET_CMD_TYPE_SET | MCM_DATE_TYPE_INTERVAL, fcinfo);
}


Datum
memcache_set_absexpire(PG_FUNCTION_ARGS) {
  return memcache_set_cmd(MCM_SET_CMD_TYPE_SET | MCM_DATE_TYPE_TIMESTAMP, fcinfo);
}


static Datum
memcache_set_cmd(int type, PG_FUNCTION_ARGS) {
  text		*key = PG_GETARG_TEXT_P(0);
  char		*val;
  size_t	 key_len, val_len;
  Interval	*span;
  float8	 expire;
  TimestampTz	 timestamptz;
  struct pg_tm	 tm;
  fsec_t	 fsec;
  u_int16_t	 flags;
  int		 ret = 0;

  MCM_CHECK(PG_RETURN_BOOL(false));

  if (PG_ARGISNULL(0))
    elog(ERROR, "unable to have a NULL key");

  key_len = VARSIZE(key) - VARHDRSZ;
  if (key_len < 1)
    elog(ERROR, "unable to have a zero length key");

  if (PG_ARGISNULL(1)) {
    val = NULL;
    val_len = 0;
  } else {
    text *val_t = PG_GETARG_TEXT_P(1);
    val = (char *) VARDATA(val_t);
    val_len = VARSIZE(val_t) - VARHDRSZ;
  }

  expire = 0.0;
  if (PG_NARGS() >= 3 && PG_ARGISNULL(2) == false) {
    if (type & MCM_DATE_TYPE_INTERVAL) {
      span = PG_GETARG_INTERVAL_P(2);
#ifdef HAVE_INT64_TIMESTAMP
      expire = (span->time / 1000000e0);
#else
      expire = span->time;
#endif
      if (span->month != 0) {
        expire += ((365.25 * 86400) * (span->month / 12));
        expire += ((30.0 * 86400) * (span->month % 12));
      }
    } else if (type & MCM_DATE_TYPE_TIMESTAMP) {
      timestamptz = PG_GETARG_TIMESTAMPTZ(2);

      /* convert to timestamptz to produce consistent results */
      if (timestamp2tm(timestamptz, NULL, &tm, &fsec, NULL, NULL) !=0)
        ereport(ERROR,
                (errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
                 errmsg("timestamp out of range")));

#ifdef HAVE_INT64_TIMESTAMP
      expire = ((timestamptz - SetEpochTimestamp()) / 1000000e0);
#else
      expire = timestamptz - SetEpochTimestamp();
#endif
    } else {
      elog(ERROR, "%s():%s:%u: invalid date type",
           __FUNCTION__, __FILE__, __LINE__);
    }
  }

  flags = 0;
  if (PG_NARGS() >= 4 && PG_ARGISNULL(3) == false)
    flags = PG_GETARG_INT16(3);

  if (type & MCM_SET_CMD_TYPE_ADD)
    ret = mcm_add(ctxt, mc, VARDATA(key), key_len, val, val_len, (time_t)expire, flags);
  else if (type & MCM_SET_CMD_TYPE_REPLACE)
    ret = mcm_replace(ctxt, mc, VARDATA(key), key_len, val, val_len, (time_t)expire, flags);
  else if (type & MCM_SET_CMD_TYPE_SET)
    ret = mcm_set(ctxt, mc, VARDATA(key), key_len, val, val_len, (time_t)expire, flags);
  else
    elog(ERROR, "%s():%s:%u\tunknown set type 0x%x",
         __FUNCTION__, __FILE__, __LINE__, type);

  if (ret == 0)
    PG_RETURN_BOOL(true);
  else if (ret > 0)
    PG_RETURN_BOOL(false);
  else
    elog(ERROR, "internal libmemcache(3) error");

  /* Not reached */
  abort();
}


Datum
memcache_server_add(PG_FUNCTION_ARGS) {
  text          *server = PG_GETARG_TEXT_P(0);
  char          *port_data;
  size_t         port_data_len;
  int            ret;
  MemoryContext  old_cxt;

  MCM_CHECK(PG_RETURN_BOOL(false));

  /* If no port specified, use the libmemcache default */
  port_data = NULL;
  port_data_len = 0;
  if (PG_NARGS() > 1) {
    text *port = PG_GETARG_TEXT_P(1);
    port_data = VARDATA(port);
    port_data_len = VARSIZE(port) - VARHDRSZ;
  }

  /* Ensure new servers are allocated in the top memory context */
  old_cxt = MemoryContextSwitchTo(TopMemoryContext);

  ret = mcm_server_add2(ctxt, mc,
                        VARDATA(server), VARSIZE(server) - VARHDRSZ,
                        port_data, port_data_len);

  MemoryContextSwitchTo(old_cxt);

  if (ret < 0) {
    elog(NOTICE, "unable to add memcache server: %d", ret);
    PG_RETURN_BOOL(false);
  }

  PG_RETURN_BOOL(true);
}


static text *
memcache_gen_host(const struct memcache_server *ms) {
  text   *ret;
  size_t  len;
  char   *cp;

  len = strlen(ms->hostname) + strlen(ms->port) + 1; /* include the colon */
  ret = (text *) palloc(len + VARHDRSZ);
  VARATT_SIZEP(ret) = len + VARHDRSZ;
  cp = VARDATA(ret);

  /* Copy the hostname */
  memcpy(cp, ms->hostname, strlen(ms->hostname));
  cp += strlen(ms->hostname);

  /* Add a colon */
  cp[0] = ':';
  cp++;

  /* Copy the port */
  memcpy(cp, ms->port, strlen(ms->port));

  return ret;
}


Datum
memcache_server_find(PG_FUNCTION_ARGS) {
  text                   *key = PG_GETARG_TEXT_P(0);
  size_t                  key_len;
  u_int32_t               hash;
  struct memcache_server *ms;
  text                   *ret;

  MCM_CHECK(PG_RETURN_BOOL(false));

  key_len = VARSIZE(key) - VARHDRSZ;
  if (key_len < 1)
    elog(ERROR, "unable to have a zero length key");

  hash = mcm_hash_key(ctxt, VARDATA(key), key_len);
  ms = mcm_server_find(ctxt, mc, hash);
  ret = memcache_gen_host(ms);

  PG_RETURN_TEXT_P(ret);
}


Datum
memcache_server_find_hash(PG_FUNCTION_ARGS) {
  struct memcache_server *ms;
  text *ret;

  MCM_CHECK(PG_RETURN_BOOL(false));

  ms = mcm_server_find(ctxt, mc, PG_GETARG_UINT32(0));
  ret = memcache_gen_host(ms);

  PG_RETURN_TEXT_P(ret);
}


Datum
memcache_stats(PG_FUNCTION_ARGS) {
  struct memcache_server_stats *stats;
  StringInfoData	            buf;

  MCM_CHECK(PG_RETURN_NULL());

  stats = mcm_stats(ctxt, mc);
  if (stats == NULL)
    elog(ERROR, "unable to create a new memcache stats structure");

  initStringInfo(&buf);
  appendStringInfo(&buf, "pid: %u\n", stats->pid);
  appendStringInfo(&buf, "uptime: %lu\n", (unsigned long) stats->uptime);
  appendStringInfo(&buf, "time: %lu\n", (unsigned long) stats->time);
  appendStringInfo(&buf, "version: %s\n", stats->version);
  appendStringInfo(&buf, "rusage_user: %lu.%ld\n",
                   (unsigned long) stats->rusage_user.tv_sec,
                   (long) stats->rusage_user.tv_usec);
  appendStringInfo(&buf, "rusage_system: %lu.%ld\n",
                   (unsigned long) stats->rusage_system.tv_sec,
                   (long) stats->rusage_system.tv_usec);
  appendStringInfo(&buf, "curr_items: %u\n", stats->curr_items);
  appendStringInfo(&buf, "total_items: %" PRIu64 "\n", stats->total_items);
  appendStringInfo(&buf, "bytes: %" PRIu64 "\n", stats->bytes);
  appendStringInfo(&buf, "curr_connections: %u\n", stats->curr_connections);
  appendStringInfo(&buf, "total_connections: %" PRIu64 "\n", stats->total_connections);
  appendStringInfo(&buf, "connection_structures: %u\n", stats->connection_structures);
  appendStringInfo(&buf, "cmd_get: %" PRIu64 "\n", stats->cmd_get);
  appendStringInfo(&buf, "cmd_set: %" PRIu64 "\n", stats->cmd_set);
  appendStringInfo(&buf, "get_hits: %" PRIu64 "\n", stats->get_hits);
  appendStringInfo(&buf, "get_misses: %" PRIu64 "\n", stats->get_misses);
  appendStringInfo(&buf, "bytes_read: %" PRIu64 "\n", stats->bytes_read);
  appendStringInfo(&buf, "bytes_written: %" PRIu64 "\n", stats->bytes_written);
  appendStringInfo(&buf, "limit_maxbytes: %" PRIu64 "\n", stats->limit_maxbytes);

  PG_RETURN_DATUM(DirectFunctionCall1(textin, CStringGetDatum(buf.data)));
}


Datum
memcache_stat(PG_FUNCTION_ARGS) {
  text		                   *stat = PG_GETARG_TEXT_P(0);
  char                         *stat_str;
  size_t                        stat_len;
  struct memcache_server_stats *stats;
  StringInfoData	            buf;

  MCM_CHECK(PG_RETURN_NULL());

  stat_str = DatumGetCString(DirectFunctionCall1(textout,
                                                 PointerGetDatum(stat)));
  stat_len = strlen(stat_str);
  if (stat_len < 1)
    elog(ERROR, "unable to have a zero length stat");

  stats = mcm_stats(ctxt, mc);
  if (stats == NULL)
    elog(ERROR, "unable to create a new memcache stats structure");

  initStringInfo(&buf);

  if (pg_strncasecmp("pid", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%u", stats->pid);
  else if (pg_strncasecmp("uptime", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%lu", (unsigned long) stats->uptime);
  else if (pg_strncasecmp("time", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%lu", (unsigned long) stats->time);
  else if (pg_strncasecmp("version", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%s", stats->version);
  else if (pg_strncasecmp("rusage_user", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%lu.%ld",
                     (unsigned long) stats->rusage_user.tv_sec,
                     (long) stats->rusage_user.tv_usec);
  else if (pg_strncasecmp("rusage_system", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%lu.%ld",
                     (unsigned long) stats->rusage_system.tv_sec,
                     (long) stats->rusage_system.tv_usec);
  else if (pg_strncasecmp("curr_items", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%u", stats->curr_items);
  else if (pg_strncasecmp("total_items", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->total_items);
  else if (pg_strncasecmp("bytes", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->bytes);
  else if (pg_strncasecmp("curr_connections", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%u", stats->curr_connections);
  else if (pg_strncasecmp("total_connections", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->total_connections);
  else if (pg_strncasecmp("connection_structures", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%u", stats->connection_structures);
  else if (pg_strncasecmp("cmd_get", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->cmd_get);
  else if (pg_strncasecmp("cmd_set", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->cmd_set);
  else if (pg_strncasecmp("get_hits", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->get_hits);
  else if (pg_strncasecmp("get_misses", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->get_misses);
  else if (pg_strncasecmp("bytes_read", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->bytes_read);
  else if (pg_strncasecmp("bytes_written", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->bytes_written);
  else if (pg_strncasecmp("limit_maxbytes", stat_str, stat_len) == 0)
    appendStringInfo(&buf, "%" PRIu64, stats->limit_maxbytes);
  else
    elog(ERROR, "unknown memcache statistic %s", stat_str);

  PG_RETURN_DATUM(DirectFunctionCall1(textin, CStringGetDatum(buf.data)));
}
